

#ifndef FONTGEN_H
#define FONTGEN_H

#define DEBUG

#ifdef DEBUG
#define DBG(...) printf("DBG:" __VA_ARGS__)
#else
#define DBG(...)
#endif


#ifdef BUILD_LITTLE_ENDIAN
#define PL_POINT_DEFINED
typedef union {
  struct{
    unsigned short p_y : 8;
    unsigned short p_x : 7;
    unsigned short p_lift: 1;
  };
  unsigned short p_data: 16;
} pl_point;
#endif

#ifdef BUILD_BIG_ENDIAN
#define PL_POINT_DEFINED
typedef union {
  struct{
    unsigned short p_lift: 1;
    unsigned short p_x : 7;
    unsigned short p_y : 8;
  };
  unsigned short p_data: 16;
} pl_point;
#endif

typedef struct {
  int gl_number;
  int gl_size;
  int gl_offset;
  char * gl_label;
  char * gl_comment;
  pl_point  * gl_points;
} pl_glyph;

typedef struct {
  char * f_label;
  int f_size; /* Number of glyphs */
  int f_base;
  int f_top;
  int f_offset;   /* Y offset to next line */
  char * f_comment;
  pl_glyph * f_glyphs;
} pl_font;

#define POINT(lift,x,y) {.p_data=0xffff &( (lift<<15)+(x<<8)+(y))}  


extern pl_font hershey_fonts;
extern pl_font hp_convert(char ** fontdata, int scale);

extern void op_font (pl_font *);

extern pl_glyph find_glyph(int);
extern pl_font assemble_font(int first_id, int size, int glyph_ids[]);

extern char* charset0[];
extern char* charset1[];
extern char* charset2[];
extern char* charset3[];
extern char* charset4[];
extern char* charset5[];
extern char* charset6[];
extern char* charset7[];
extern char* charset173[];
extern char* charset205[];

extern void font_init(void);
extern pl_font font_cyrilc;
extern pl_font font_gothgbt;
extern pl_font font_gothgrt;
extern pl_font font_gothitt;
extern pl_font font_greekc;
extern pl_font font_greekcs;
extern pl_font font_greekp;
extern pl_font font_greeks;
extern pl_font font_italicc;
extern pl_font font_italiccs;
extern pl_font font_italict;
extern pl_font font_romanc;
extern pl_font font_romancs;
extern pl_font font_romand;
extern pl_font font_romanp;
extern pl_font font_romans;
extern pl_font font_romant;
extern pl_font font_scriptc;
extern pl_font font_scripts;



extern pl_font font_astrology;
extern pl_font font_cursive;
extern pl_font font_cyrilc_1;
extern pl_font font_cyrillic;
extern pl_font font_futural;
extern pl_font font_futuram;
extern pl_font font_gothgbt;
extern pl_font font_gothgrt;
extern pl_font font_gothiceng;
extern pl_font font_gothicger;
extern pl_font font_gothicita;
extern pl_font font_gothitt;
extern pl_font font_greekc;
extern pl_font font_greek;
extern pl_font font_greeks;
extern pl_font font_japanese;
extern pl_font font_markers;
extern pl_font font_mathlow;
extern pl_font font_mathupp;
extern pl_font font_meteorology;
extern pl_font font_music;
extern pl_font font_rowmand;
extern pl_font font_rowmans;
extern pl_font font_rowmant;
extern pl_font font_scriptc;
extern pl_font font_scripts;
extern pl_font font_symbolic;
extern pl_font font_timesg;
extern pl_font font_timesib;
extern pl_font font_timesi;
extern pl_font font_timesrb;
extern pl_font font_timesr;

#endif
